#!/usr/bin/env python
#-*- coding: utf-8 -*-

# License: GPLv3
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY# without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# To view a copy of the GNU General Public License
# visit: http://www.gnu.org/licenses/gpl.html

# ------------
#| Change Log |
# ------------
# Older version Rel.1-7: Actions RPL http://gimpchat.com/viewtopic.php?f=9&t=17395

# Ver. 1.0: Initial release: MrQ Actions RPL (09/10/2021)
# Ver. 1.1: (12/10/2021) replace: File to Load: (file selector) to Actions Presets: (select an action from the list)
# Ver. 1.2: (29/10/2021) add options: Color-> Remove Color; Source & Target -> Shuffle Layers (random) Selections -> Fill in Color/Pattern
#                        bug fix: If the file 'GimpCurvesConfig.settings' (in directory: filters) or the folder 'Actions-Presets' (in directory: plug-ins) does not exist
#                        - the plugin does not register in the menu or displays an error message when using >>> now they will be created. 

# Codes used:
# run_code_on_visible_layers.py  by code Tin Tran: https://gimplearn.net/viewtopic.php?f=3&t=1027
# apply_curve by code Tin Tran: http://gimpchat.com/viewtopic.php?f=9&t=14058&start=0
# apply_color_curves_preset.py by code Tin Tran: http://gimpchat.com/viewtopic.php?f=9&t=14064&start
# ofn-export-layers.py by code Ofnuts:  https://sourceforge.net/projects/gimp-tools/files/scripts/ 
# gimfu3 with tabs by J.F.Garcia:  https://www.arakne.es/en/gimp-plugin-follow-path-map-values-on-the-y-coordinate/
# gegl_command by Kevin Payne & Claude Lion [cli345]: http://gimpchat.com/viewtopic.php?f=9&t=18040&start=30#p264354 

# If you want to use all the features of the plugin, you should add:

# A. Old Plug-ins (added to Gimp 2.10) from Gimp 2.8 (download https://www.gimpscripts.net/2020/11/filters-from-gimp-28-that-have-become.html)

# 1. alien-map.exe → Colors/Map/Alien Map...
# 2. apply-canvas.exe → Filters/Artistic/Apply Canvas...
# 3. blur-gauss.exe → Filters/Blur/Gaussian Blur...
# 4. blur-gauss-selective.exe → Filters/Blur/Selective Gaussian Blur...
# 5. blur-motion.exe → Filters/Blur/Motion Blur...
# 6. bump-map.exe → Filters/Map/Bump Map...
# 7. channel-mixer.exe → Colors/Components/Channel Mixer...
# 8. color-rotate.exe → Colors/Map/Rotate Colors...
# 9. color-to-alpha.exe → Layer/Transparency/Color to Alpha... and Colors/Color to Alpha...
#10  convolution-matrix.exe → Filters/Generic/Convolution Matrix...
#11. cubism.exe → Filters/Artistic/Cubism...
#12. deinterlace.exe → Filters/Enhance/Deinterlace...
#13. displace.exe → Filters/Map/Displace...
#14. edge.exe → Filters/Edge-Detect/Edge...
#15. edge-laplace.exe → Filters/Edge-Detect/Laplace...
#16. edge-neon.exe → Filters/Edge-Detect/
#17. edge-sobel.exe → Filters/Edge-Detect/Sobel...
#18. engrave.exe → Filters/Distorts/Engrave...
#19. illusion.exe → Filters/Map/Illusion...
#20. lens-apply.exe → Filters/Distorts/Apply Lens...
#21. lens-distortion.exe → Filters/Distorts/Lens Distortion...
#22. maze.exe → Filters/Render/Pattern/Maze...
#23. mosaic.exe → Filters/Distorts/Mosaic...
#24. newsprint.exe → Filters/Distorts/Newsprint... (from Gimp-2.10.12)
#25. noise-hsv.exe → Filters/Noise/HSV Noise...
#26. noise-randomize.exe → Filters/Noise/Hurl...
#27. noise-rgb.exe → Filters/Noise/RGB Noise...
#28. noise-solid.exe → Filters/Render/Clouds/Solid Noise...
#29. noise-spread.exe → Filters/Noise/Spread...
#30  oilify.exe → Filters/Artistic/Oilify...
#31. pixelize.exe → Filters/Blur/Pixelize...
#32. plasma.exe → Filters/Render/Clouds/Plasma...
#33. polar-coords.exe → Filters/Distorts/Polar Coordinates...
#34. red-eye-removal.exe → Filters/Enhance/Red Eye Removal...
#35. ripple.exe → Filters/Distorts/Ripple...
#36. shift.exe → Filters/Distorts/Shift...
#37. sinus.exe → Filters/Render/Pattern/Sinus...
#38. tile-glass.exe → Filters/Artistic/Glass Tile...
#39  threshold-alpha.exe → Layer/Transparency/Threshold Alpha...
#40  unsharp-mask.exe → Filters/Enhance/Unsharp Mask...
#41. value-propagate.exe → Filters/Distorts/Value Propagate...
#42. video.exe → Filters/Distorts/Video...
#43. whirl-pinch.exe →   Filters/Distorts/Whirl and Pinch...
#44. waves.exe → Filters/Distorts/Waves...
#45. wind.exe → Filters/Distorts/Wind...

# B. Felimage Noise 0.1.1 Win_64bit https://www.gimpscripts.net/2020/09/felimage-noise-011-win64bit.html
# C. GMIC-QT Win_64bit (minimum 2.9.9 or newer) https://gmic.eu/files/windows/
# D. MathMap Win_32bit https://www.gimp-forum.net/Thread-Install-the-MathMap-plug-in-for-Windows?pid=19023#pid19023
# E. Userfilter-0.9.7 Win_64bit  https://www.gimpscripts.net/2020/10/userfilter-097-win64bit.html


############################################################################################################################################################
########################################################## START gimpfu3 by J.F.Garcia [arakne] ############################################################
############################################################################################################################################################
#   Gimp-Python - allows the writing of GIMP plug-ins in Python.
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import string as _string
import math
import gimp
import gimpcolor
import re
from gimpenums import *
pdb = gimp.pdb
import colorsys
import gettext
from gimpshelf import shelf
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)
_ = t.ugettext

class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}

def register(proc_name, blurb, help, author, copyright, date, label, imagetypes, params, results, function, menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
		return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead" % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]], x[1], _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode", "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results)
        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
    gimpui.gimp_ui_init ()  # add MrQ	
#    import pango

    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default=""):
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            #self.set_line_wrap(True)  			# disabled by MrQ
            #self.set_single_line_mode(False)	# disabled by MrQ

        def get_value(self):
            return self.get_text()

    class LabelTitle(gtk.Label):
        def __init__(self, default="", align = (0.0, 0.0)):
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default=""):
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))

    class ToggleEntry(gtk.ToggleButton):
        def __init__(self, default=0, labels=[_("No"),_("Yes")]):
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()
            self.connect("toggled", self.changed)
            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()

    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("Yes"), 1), (_("No"), 0))):
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=()):
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)

        def get_value(self):
            return self.get_active()

    def FileSelector(default=""):
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            selector = DirnameSelector
            if default == "/": default = ""
        else:
            selector = FilenameSelector
        return selector(default)

    class FilenameSelector(gtk.FileChooserButton):
        def __init__(self, default="", save_mode=False):
            gtk.FileChooserButton.__init__(self,_("Python-Fu File Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default=""):
            gtk.FileChooserButton.__init__(self,_("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : gimpui.ColorSelector,
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : gimpui.FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : gimpui.PatternSelector,
            PF_GRADIENT    : gimpui.GradientSelector,
            PF_PALETTE     : gimpui.PaletteSelector,
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    vbox = gtk.VBox(False, 12)
    vbox.set_border_width(12)
    dialog.vbox.pack_start(vbox)
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)


    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})

    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        gtk.main_quit()

    iniX=0
    dialog.connect("response", response)
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]
 
        tab, incr, rows, connect, cols, hlp = (0, 1, 1, None, 1, desc)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            hlp  = getXtra(xtra, 'help'   , desc)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE and desc != '':
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW ):
                wid = _edit_mapping[pf_type](def_val, paramsi[4])
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4])
                else:
                    wid = _edit_mapping[pf_type](def_val)
            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val)
            if pf_type!=PF_TITLE and desc != '':
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', hlp)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()

    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break
        min_args = i
    if len(script_params) > min_args:
        start_params = params[:min_args + 1]
        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE
    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])
        gimp.displays_flush()
    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message

############################################################################################################################################################
######################################################### END gimpfu3 ######################################################################################
############################################################################################################################################################

####################################################################################################################
################# START gegl_command by cli345 #####################################################################
####################################################################################################################

from sys import platform
from ctypes import *

def load_library (library_name):
	if platform == "linux" or platform == "linux2":
		library_name = library_name + '.so.0'
	elif platform == "win32":
		from ctypes.util import find_library
		library_name = find_library (library_name + "-0")
	else:
		raise BaseException ("TODO")
	return CDLL (library_name)

gimpC = load_library ('libgimp-2.0')


def gegl_batch(image, drawable, gegl_graph_string):
	
	class GeglBuffer(Structure):
		pass
	drawable_id = drawable.ID
	
	gegl = load_library ('libgegl-0.4')
	gegl.gegl_init (None, None)
	gimpC.gimp_drawable_get_shadow_buffer.restype = POINTER (GeglBuffer)
	gimpC.gimp_drawable_get_buffer.restype        = POINTER (GeglBuffer)

	x = c_int (); y = c_int (); w = c_int (); h = c_int ()
	non_empty,x,y,w,h = pdb.gimp_drawable_mask_intersect (drawable)
	args = [b"string", c_char_p (gegl_graph_string), c_void_p ()]
	
	if non_empty:
		source = gimpC.gimp_drawable_get_buffer (drawable_id)
		target = gimpC.gimp_drawable_get_shadow_buffer (drawable_id)
		gegl.gegl_render_op (source, target, "gegl:gegl", *args)
		gegl.gegl_buffer_flush (target)
		gimpC.gimp_drawable_merge_shadow (drawable_id, PushUndo = True)
		gimpC.gimp_drawable_update (drawable_id, x, y, w, h)
		gimpC.gimp_displays_flush ()
		
####################################################################################################################
################# END gegl_command by cli345 #######################################################################
####################################################################################################################		

####################################################################################################
################# START apply_curve by code Tin Tran ###############################################
####################################################################################################
import string
def readcurvefile(
  curvefilename
  ):
  curvefile = open( u''+curvefilename,'r' )
  lines = curvefile.readlines( )
  #if lines[0] == "# GIMP curves tool settings\n": # MrQ
  if lines[0] == "# GIMP 'Curves' settings\n" or "# GIMP curves tool settings":  # MrQ
    for line in lines:
      if line == "(channel value)\n": 
        nameCurve = "value"
      if line == "(channel red)\n": 
        nameCurve = "red"
      if line == "(channel green)\n": 
        nameCurve = "green"
      if line == "(channel blue)\n": 
        nameCurve = "blue"
      if line == "(channel alpha)\n": 
        nameCurve = "alpha"
      if line != "\n" :
        words = string.split( line )
        if words[0] == "(points":
          if nameCurve == "value":
            curve = words[2:]
          if nameCurve == "red":
            curvered = words[2:]
          if nameCurve == "green":
            curvegreen = words[2:]
          if nameCurve == "blue":
            curveblue = words[2:]
          if nameCurve == "alpha":
            curvealpha = words[2:]
    return curve, curvered, curvegreen, curveblue # Alpha is ignored
  else:
    pdb.gimp_message("This is not a GIMP New Curves File Format")
  curvefile.close()  

def extractpoints( curve ):
  icurve= [];
  for curvestr in curve:
    if curvestr[-1] == ")" :
      curvestr = curvestr[:-1]
    curvefloat = float( curvestr )
    curveint = (curvefloat * 255.0) #scale to [0,255]
    icurve.append( int( curveint ))
  ibmax = int( len(icurve) ) -1
  ib= 0
  ixcurve= []
  while ib <= ibmax:
    if icurve[ ib ] != -255:
      ixcurve.append( icurve[ ib ] )
      ixcurve.append( icurve[ ib+1 ] )
    ib= ib +2
  return ixcurve


def apply_curve_from_file(image,layer,curvefilename):
	
	curve, curvered, curvegreen, curveblue = readcurvefile( curvefilename ) # Alpha is ignored
	icurve= extractpoints( curve )
	icurvered= extractpoints( curvered )
	icurvegreen= extractpoints( curvegreen )
	icurveblue= extractpoints( curveblue )
  
    #apply curve information to current layer
	if len(icurve) >= 4:
		pdb.gimp_curves_spline(layer, HISTOGRAM_VALUE, len(icurve), icurve )
	if len(icurvered) >= 4:
		pdb.gimp_curves_spline(layer, HISTOGRAM_RED, len(icurvered), icurvered )
	if len(icurvegreen) >= 4:
		pdb.gimp_curves_spline(layer, HISTOGRAM_GREEN, len(icurvegreen), icurvegreen )
	if len(icurveblue) >= 4:
		pdb.gimp_curves_spline(layer, HISTOGRAM_BLUE, len(icurveblue), icurveblue )
####################################################################################################
################# END apply_curve by code Tin Tran #################################################
####################################################################################################

####################################################################################################
################# START apply_color_curves_preset.py by code Tin Tran ##############################
####################################################################################################
def local_get_color_curves_presets_names():
	curvepresetfile =  gimp.directory +  os.sep + 'filters' +  os.sep +  'GimpCurvesConfig.settings'
	if not os.path.isfile(curvepresetfile):
		f = open(curvepresetfile, "w")
		f.write('# settings')
		f.write('\n\n(GimpCurvesConfig "Curves Saved Settings Not Exist"')
		f.write('\n\n# end of settings')
		f.close()
	ret_list = []
	curvepresetfile =  gimp.directory +  os.sep + 'filters' +  os.sep +  'GimpCurvesConfig.settings'
	curvepresetfile = os.path.expanduser(curvepresetfile)
	curvefile = open( u''+curvepresetfile,'r' )
	lines = curvefile.readlines( )
	for line in lines:
		if line.find('(GimpCurvesConfig "') > -1:
			start = line.find('"') + 1
			end = line.find('"',start)
			ret_list.append(line[start:end])	
	curvefile.close()		
	return (ret_list)
	
def readcurvefile_presetid(id):
	ret_lines = []
	curvepresetfile =  gimp.directory +  os.sep + 'filters' +  os.sep +  'GimpCurvesConfig.settings'
	curvepresetfile = os.path.expanduser(curvepresetfile)
	curvefile = open( u''+curvepresetfile,'r' )
	lines = curvefile.readlines( )
	current_preset_id = -1
	found_it = False
	for line in lines:
		if found_it: #means we're still looking at the right preset_id lines
			ret_lines.append(line[4:]) #strip away the first 4 spaces 
		if line.find('(GimpCurvesConfig "') > -1:
			current_preset_id += 1
		if current_preset_id == id:
			found_it = True
		else:
			found_it = False
	curvefile.close()
	return (ret_lines)
def readcurvepresetfile(colorcurvepresetid):
	lines = readcurvefile_presetid(colorcurvepresetid)
	for line in lines:
		if line == "(channel value)\n": 
			nameCurve = "value"
		if line == "(channel red)\n": 
			nameCurve = "red"
		if line == "(channel green)\n": 
			nameCurve = "green"
		if line == "(channel blue)\n": 
			nameCurve = "blue"
		if line == "(channel alpha)\n": 
			nameCurve = "alpha"

		if line != "\n" and line.find("(") > -1:
			words = string.split( line )
			if words[0] == "(samples": #Tin Tran use (samples instead of (points because (points didn't work for freehand curves)
				if nameCurve == "value":
					curve = words[2:]
				if nameCurve == "red":
					curvered = words[2:]
				if nameCurve == "green":
					curvegreen = words[2:]
				if nameCurve == "blue":
					curveblue = words[2:]
				if nameCurve == "alpha":
					curvealpha = words[2:]
	return curve, curvered, curvegreen, curveblue, curvealpha
  

def extractpoints_( curve ):
	icurve= [];
	for curvestr in curve:
		curvefloat = float( curvestr.rstrip(")") )
		curveint = (curvefloat * 255.0) #scale to [0,255]
		icurve.append( int( curveint ))
	return icurve

def apply_color_curves_from_preset(image,layer,presetid):

	curve, curvered, curvegreen, curveblue, curvealpha = readcurvepresetfile( presetid ) #Tin Tran not sure why alpha was previously ignored.
	icurve= extractpoints_( curve )
	icurvered= extractpoints_( curvered )
	icurvegreen= extractpoints_( curvegreen )
	icurveblue= extractpoints_( curveblue )
	icurvealpha = extractpoints_( curvealpha )  #Tin tran not sure why alpha was ignored i am adding it in

	pdb.gimp_curves_explicit(layer, HISTOGRAM_VALUE, len(icurve), icurve )
	if not pdb.gimp_item_is_layer_mask(layer): #if it's not layer mask try to apply all other channels
		pdb.gimp_curves_explicit(layer, HISTOGRAM_RED, len(icurvered), icurvered )
		pdb.gimp_curves_explicit(layer, HISTOGRAM_GREEN, len(icurvegreen), icurvegreen )
		pdb.gimp_curves_explicit(layer, HISTOGRAM_BLUE, len(icurveblue), icurveblue )
		if pdb.gimp_drawable_has_alpha(layer): #if layer has alpha we apply alpha.
			pdb.gimp_curves_explicit(layer, HISTOGRAM_ALPHA, len(icurvealpha), icurvealpha )
			
####################################################################################################
################# END apply_color_curves_preset.py by code Tin Tran ################################
####################################################################################################
import random
def shuffle_layers(image,layer):

	num_layers, layer_ids = pdb.gimp_image_get_layers(image)
	layer_names_list = image.layers
	loop = 0
	shuffle = num_layers - 1
	if shuffle > 1:
		while loop < num_layers -1:
			random.seed()
			iloop = 0
			deal = random.randrange(0,num_layers -2)
			random.seed()
			pdb.gimp_image_lower_item(image, layer_names_list[deal])
			num_layers, layer_ids = pdb.gimp_image_get_layers(image)
			layer_names_list = image.layers
			loop += 1
			
import os.path
def save_code(image, layer, code, name_preset):
	codem = str(code)

		
	if not os.path.exists(gimp.directory +  os.sep + 'plug-ins' +  os.sep + 'Actions-Presets'):
		os.mkdir(gimp.directory +  os.sep + 'plug-ins' +  os.sep + 'Actions-Presets')
	
	backup_filefullpath = gimp.directory +  os.sep + 'plug-ins' +  os.sep + 'Actions-Presets' +  os.sep + str(name_preset) + '.txt'
	f = open(backup_filefullpath,"w")
	f.write(codem)	
	f.close()
	
def rgb_to_hex(rgb_color):
	r = rgb_color[0]
	g = rgb_color[1]
	b = rgb_color[2]
	return '#%02x%02x%02x' % (r,g,b)


import math, glob
import sys,os,os.path,traceback
global apply_layers
apply_layers = []
def actions_path():
	if not os.path.exists(gimp.directory +  os.sep + 'plug-ins' +  os.sep + 'Actions-Presets'):
		os.mkdir(gimp.directory +  os.sep + 'plug-ins' +  os.sep + 'Actions-Presets')
	actionspresetfile =  gimp.directory +  os.sep + 'plug-ins' +  os.sep + 'Actions-Presets' +  os.sep
	glob.glob(actionspresetfile)
	return os.listdir(actionspresetfile)
	
def check_layers(image,layer,layergroup):

	if layergroup==0: # All layers
		apply_layers.append(layer)

	elif layergroup==1: # Visible layers
		if layer.visible:
			apply_layers.append(layer)
			
	elif layergroup==2: #Not visible layers
		if not layer.visible:
			apply_layers.append(layer)
			
	elif layergroup==3: # Linked layers
		if layer.linked:
			apply_layers.append(layer)
			
	elif layergroup==4: # Not-linked layers
		if not layer.linked:
			apply_layers.append(layer)
			
	elif layergroup==5: # Active only
		#image = pdb.gimp_item_get_image(layer)
		x = pdb.gimp_image_get_active_layer(image)
		layer_position = pdb.gimp_image_get_item_position(image,x)
		#for x in image.layers:
		position = pdb.gimp_image_get_item_position(image,layer)
		if position == layer_position:
			apply_layers.append(layer)
				
	elif layergroup==6: # Active and Above layers
		#image = pdb.gimp_item_get_image(layer)
		x = pdb.gimp_image_get_active_layer(image)
		layer_position = pdb.gimp_image_get_item_position(image,x)
		#for x in image.layers:
		position = pdb.gimp_image_get_item_position(image,layer)
		if position <= layer_position:
			apply_layers.append(layer)
				
	elif layergroup==7: # Active and Below layers
		#image = pdb.gimp_item_get_image(layer)
		x = pdb.gimp_image_get_active_layer(image)
		layer_position = pdb.gimp_image_get_item_position(image,x)
		#for x in image.layers:
		position = pdb.gimp_image_get_item_position(image,layer)
		if position >= layer_position:
			apply_layers.append(layer)
							
	for child in layer.children: # recursive call to check on children here.
		check_layers(image,child,layergroup)
		
#================= Start Export Layer by code (c) Ofnuts =======================================================		
def exportLayers_arpl(image,layer,directory,namePattern):

    try:
        if not image.filename:
            imageName='Layer'
        else:
            imageName,_=os.path.splitext(os.path.basename(image.filename))
        layersCount=len(image.layers)
        formatValues={}
        formatValues['imageName']=imageName
        formatValues['count']=layersCount
        # for i,layer in enumerate(image.layers): MrQ no all layers only " Choose layers: "	
        for i,layer in enumerate(apply_layers):   # rename image.layers to apply_layers
            formatValues['numDown0']=i
            formatValues['numDown1']=i+1
            formatValues['numUp0']=layersCount-(i+1)
            formatValues['numUp1']=layersCount-i
            formatValues['rawName']=layer.name
            formatValues['name']=os.path.splitext(layer.name)[0]
            formatValues['width']=layer.width
            formatValues['height']=layer.height
    
            try:
                filename=namePattern.format(**formatValues)
            except KeyError as e:
                raise Exception('No formatting variable called "%s"' % e.args[0])        
            filename=os.path.join(directory,filename)
            pdb.gimp_file_save(image,layer,filename,filename)
            
    except Exception as e:
        pdb.gimp_message(e.args[0])
        print traceback.format_exc()
# ============== End Export Layer by code (c) Ofnuts ===============================================================================

# Copy multiple times iterations=Number of copies		
def copy_layers_RPL(image, drawable, iterations):  

	for item in range(0,int(iterations)):
		copy=pdb.gimp_layer_copy(drawable,1)
		pdb.gimp_image_add_layer(image,copy,2)

	pdb.gimp_displays_flush()
	
def python_fu_MrQ_Actions_Run_Plugin_on_Layers(image, layer, layergroup, rev,mode, xopacity,opacity,
	cocu1,cocu2,cocu3 ,cocu4,cocu5,iter,cocu6,cocu7,
	mir1,mir2,mir3,mir4,mir5,mir6,mir7,mir8,mir9,mir10,
	trans1,trans2,trans3,trans4,trans5,trans6,trans7,trans8,trans9,
	color_info,color_p, name_p,color_f, name_f,colorize, color_hsl,color_rem, color_rgb,
	scale,interpolation,local_origin, percent,new_width, new_height,offset, offset_x, offset_y,
	sel_info,sel1,sel2,sel3,color_fg,sel4,sel_pattern,
	plugin_info,vals, plugin,gegl_info,Op,
	code_r,code_info,code,code_save,code_info_name,name_preset,load_info,load_name_actions,	
	canva1,canva2,canva3,
	rename, export, directory, namePattern,ptab, help):

	pdb.gimp_image_undo_group_start(image)
	pdb.gimp_context_push()
	#name = pdb.gimp_item_get_name(layer)
	
	preset = ""
	for currentlayer in image.layers:
		check_layers(image,currentlayer, layergroup)
	
	# Run-mode	
	
	if vals==0:
		vals=RUN_INTERACTIVE
	if vals==1:
		vals=RUN_WITH_LAST_VALS
	# run_mode=RUN_NONINTERACTIVE
    
	if rev==5:	# GROUP

		group_new = pdb.gimp_layer_group_new(image)
		parent = pdb.gimp_item_get_parent(group_new)
		position = pdb.gimp_image_get_item_position(image, layer)
		pdb.gimp_image_insert_layer(image,group_new, None, 0) 
		group_new.name = "Group New"
		for layer in apply_layers:
			pdb.gimp_image_reorder_item(image, layer, group_new, position)
			pdb.gimp_image_lower_item_to_bottom(image, layer)
			
	if cocu1: # COPY & PASTE TOP to DOWN
	
		pdb.gimp_image_set_active_layer(image,layer)
		for layer in reversed(apply_layers):
			copy = pdb.gimp_layer_copy(layer, TRUE)
			pdb.gimp_image_insert_layer(image, copy, None, 0)
			
	if cocu6:  # #  start OPEN AS LAYERS IN NEW IMAGE cd..
		image_create = pdb.gimp_image_new(image.width, image.height, 0)
		display = pdb.gimp_display_new(image_create)
					
	if trans5 or trans6:
		parent = pdb.gimp_item_get_parent(layer)
		name = pdb.gimp_item_get_name(layer)
		position = pdb.gimp_image_get_item_position(image , layer)
		pdb.gimp_edit_copy(layer)
		copy = pdb.gimp_edit_paste(layer ,1)
		pdb.gimp_floating_sel_to_layer(copy)
		pdb.gimp_image_raise_item_to_top(image, copy)

	if sel2:
		pdb.gimp_selection_save(image)

	for layer in apply_layers:
		#pdb.gimp_image_set_active_layer(image,layer)
			
		if rev==1:	# VISIBLE LAYER
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_item_set_visible(layer, 1)			
		if rev==2: # NOT-VISIBLE LAYER
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_item_set_visible(layer, 0)			
		if rev==3:	# LINKED LAYER
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_item_set_linked(layer, 1)				
		if rev==4:	# NOT-LINKED LAYER
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_item_set_linked(layer, 0)
			
		if rev==6:	# GROUPS IND.
			group_new = pdb.gimp_layer_group_new(image)
			parent = pdb.gimp_item_get_parent(group_new)
			position = pdb.gimp_image_get_item_position(image, layer)
			pdb.gimp_image_insert_layer(image,group_new, None, 0) 
			group_new.name = "Group New"
			#for layer in apply_layers:
			pdb.gimp_image_reorder_item(image, layer, group_new, position)
			pdb.gimp_image_lower_item_to_bottom(image, layer)
			
		if rev==7:	# UN-GROUP
			pdb.gimp_image_set_active_layer(image,layer)		
			position = pdb.gimp_image_get_item_position(image, layer)
			for child_layer in reversed(layer.children):
				pdb.gimp_image_reorder_item(image, child_layer, layer.parent, position)
			if pdb.gimp_item_is_group(layer):
				pdb.gimp_image_remove_layer(image, layer)
		if rev==8: #RAISES LAYER TO TOP (\"Reverse\"
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_image_raise_item_to_top(image, layer)
		if rev==9:	# RAISES LAYER TO BOTTOM (\"Reverse\"
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_image_lower_item_to_bottom(image, layer)
		if rev==10:  # CLEAR LAYERS
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_edit_clear(layer)		
		if rev==11:	# REMOVE LAYERS
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_image_remove_layer(image, layer)
		if rev==12: # Revers Visiible
			pdb.gimp_image_set_active_layer(image,layer)
			if layer.visible:		
				pdb.gimp_item_set_visible(layer, 0)
			else:				
				pdb.gimp_item_set_visible(layer, 1)			
		if rev==13:
			pdb.gimp_image_set_active_layer(image,layer)
			shuffle_layers(image,layer)
		if cocu2: # COPY & PASTE REVERSED
			pdb.gimp_image_set_active_layer(image,layer)
			copy = pdb.gimp_layer_copy(layer, TRUE)
			pdb.gimp_image_insert_layer(image, copy, None, 0)
		if cocu3: # COPY & PASTE ABOVE CHOOSEN LAYERS 
			pdb.gimp_image_set_active_layer(image,layer)
			new_layer = pdb.gimp_layer_new_from_drawable(layer, image)
			pdb.gimp_image_insert_layer(image,new_layer,None,-1)

		if cocu4: # COPY & PASTE below CHOOSEN LAYERS 
			pdb.gimp_image_set_active_layer(image,layer)
			new_layer = pdb.gimp_layer_new_from_drawable(layer, image)
			pdb.gimp_image_insert_layer(image,new_layer,None,-1)
			pdb.gimp_image_lower_item(image, new_layer)			
		if cocu5:	#		
			pdb.gimp_image_set_active_layer(image,layer)
			copy_layers_RPL(image,layer,iter)

		if canva1:	# " AUTOCROP 
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_autocrop_layer(image, layer)		
		if canva2:	# LAYERS TO IMAGE SIZE
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_layer_resize_to_image_size(layer)			
		if canva3:	# IMAGE RESIZE TO LAYERS
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_image_resize_to_layers(image)

			
		if cocu6:     # cd OPEN AS LAYERS IN NEW IMAGE
			pdb.gimp_image_set_active_layer(image,layer)		
			item_copy = pdb.gimp_layer_new_from_drawable(layer, image_create)
			pdb.gimp_image_insert_layer(image_create, item_copy, None, 0)
			item_copy.name = layer.name
			pdb.gimp_drawable_update(item_copy, 0, 0, item_copy.width, item_copy.height)
			pdb.gimp_image_lower_item_to_bottom(image_create, item_copy)
		if cocu7:		 # OPEN LAYERS AS NEW IMAGES
			pdb.gimp_image_set_active_layer(image,layer)
			image_create = pdb.gimp_image_new(image.width, image.height, 0)
			display = pdb.gimp_display_new(image_create)
			item_copy = pdb.gimp_layer_new_from_drawable(layer, image_create)
			pdb.gimp_image_insert_layer(image_create, item_copy, None, 0)
			item_copy.name = layer.name
			pdb.gimp_drawable_update(item_copy, 0, 0, item_copy.width, item_copy.height)

		if mir1:	# " CENTERED "
			wL = pdb.gimp_drawable_width(layer)
			hL = pdb.gimp_drawable_height(layer)
			wI = pdb.gimp_image_width(image)
			hI = pdb.gimp_image_height(image)
			pdb.gimp_image_set_active_layer(image,layer)

			sw=(wI-wL)/2
			sh=(hI-hL)/2		
			pdb.gimp_layer_set_offsets(layer, sw, sh)
			
			
		if mir2: # " ALIGN TOP "
			pdb.gimp_image_set_active_layer(image,layer)
			a,b = pdb.gimp_drawable_offsets(layer)
			pdb.gimp_layer_set_offsets (layer,a,0)
			
		if mir3: # " ALIGN BOTTOM "
			pdb.gimp_image_set_active_layer(image,layer)
			a,b = pdb.gimp_drawable_offsets(layer)
			pdb.gimp_layer_set_offsets (layer, a ,image.height - layer.height) 


		if mir4: # " ALIGN RIGHT "		
			pdb.gimp_image_set_active_layer(image,layer)
			a,b = pdb.gimp_drawable_offsets(layer)			
			pdb.gimp_layer_set_offsets (layer,image.width-layer.width, b)
			
		if mir5: # " ALIGN LEFT "
			pdb.gimp_image_set_active_layer(image,layer)	
			a,b = pdb.gimp_drawable_offsets(layer)
			pdb.gimp_layer_set_offsets (layer, 0, b)
			
		if mir6: # " FLIP HORIZONTAL "u'\u21e8'"\n (Don't use Linked Layer)"
			wL = pdb.gimp_drawable_width(layer)
			pdb.gimp_selection_none(image)
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.gimp_item_transform_flip_simple(layer, 0, FALSE, wL/2)
		if mir7: # " FLIP VERTICAL "u'\u21e9'"\n (Don't use Linked Layer)"
			hL = pdb.gimp_drawable_height(layer)
			pdb.gimp_selection_none(image)
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.gimp_item_transform_flip_simple(layer, 1, FALSE, hL/2)
		if mir8: # " ROTATE-90° "
			wL = pdb.gimp_drawable_width(layer)
			hL = pdb.gimp_drawable_height(layer)
			pdb.gimp_selection_none(image)			
			pdb.gimp_item_transform_rotate_simple(layer, 0, FALSE, wL/2, hL/2)
		if mir9: # " ROTATE-180° "
			wL = pdb.gimp_drawable_width(layer)
			hL = pdb.gimp_drawable_height(layer)
			pdb.gimp_selection_none(image)			
			pdb.gimp_item_transform_rotate_simple(layer, 1, FALSE, wL/2, hL/2)
		if mir10: # " ROTATE-270° "
			wL = pdb.gimp_drawable_width(layer)
			hL = pdb.gimp_drawable_height(layer)
			pdb.gimp_selection_none(image)	
			pdb.gimp_item_transform_rotate_simple(layer, 2, FALSE, wL/2, hL/2)

			
		if trans1:
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_layer_add_alpha(layer)
		if trans2:
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_layer_flatten(layer)	
		if trans3:
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_layer_set_lock_alpha(layer, 1)
		if trans4:
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_layer_set_lock_alpha(layer, 0)
		if trans5: 	# OVERLAY (active layer on chosen)
			pdb.gimp_item_set_visible(layer, 1)	
			parent = pdb.gimp_item_get_parent(layer)
			position = pdb.gimp_image_get_item_position(image, layer)
			new_layer = pdb.gimp_layer_new_from_drawable(copy, image)
			pdb.gimp_image_insert_layer(image,new_layer,parent,position)
			pdb.gimp_image_merge_down(image, new_layer, EXPAND_AS_NECESSARY)
			#pdb.gimp_image_merge_down(image, new_layer,CLIP_TO_BOTTOM_LAYER)
		if trans6: # COMBINE\n (active layer under the chosen)
			pdb.gimp_item_set_visible(layer, 1)	
			parent = pdb.gimp_item_get_parent(layer)
			position = pdb.gimp_image_get_item_position(image, layer)
			new_layer = pdb.gimp_layer_new_from_drawable(copy, image)
			pdb.gimp_image_insert_layer(image,new_layer,parent,position)
			pdb.gimp_image_raise_item(image, layer)
			pdb.gimp_image_merge_down(image, layer, EXPAND_AS_NECESSARY)
			#pdb.gimp_image_merge_down(image, new_layer,CLIP_TO_BOTTOM_LAYER)			
		if trans7:	# alpha_to_path
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_image_select_item(image, 2, layer)		
			pdb.plug_in_sel2path(image, layer)
			pdb.gimp_selection_none(image)
		if trans8:	# Alpha to Selection and Crop
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.gimp_image_select_item(image, 2, layer)
			#pdb.plug_in_autocrop_layer(image, layer)
			pdb.gimp_selection_none(image)
		if trans9:	# semiflatten			
			pdb.plug_in_semiflatten(image,layer)
			
		if color_p:
			apply_color_curves_from_preset(image,layer,name_p)
		
		if color_f:
			apply_curve_from_file(image,layer,name_f)
			
		if colorize:
			pdb.gimp_image_set_active_layer(image,layer)
			
			h,l,s = colorsys.rgb_to_hls(color_hsl[0]/255.0,color_hsl[1]/255.0,color_hsl[2]/255.0)
			hue = h * 360.0
			saturation = s * 100.0
			lightness = l * 200.0 - 100.0
			pdb.gimp_drawable_colorize_hsl(layer, hue, saturation, lightness)

		if color_rem:
            #pdb.gimp_layer_add_alpha(overlay)
			pdb.gimp_context_set_antialias(1)   # False = 0 i True = 1
			pdb.gimp_image_select_color(image, 2, layer, (color_rgb))
			pdb.gimp_selection_grow(image,1)
            #pdb.gimp_selection_bounds(image)
			pdb.gimp_drawable_edit_clear(layer)
			pdb.gimp_selection_none(image)			
		
		
		if scale==1 or scale==2:
			pdb.gimp_context_set_interpolation(interpolation)	# INTERPOLATION-NONE (0), INTERPOLATION-LINEAR (1), INTERPOLATION-CUBIC (2), INTERPOLATION-NOHALO (3), INTERPOLATION-LOHALO (4)		
			
		if scale==1:
			pdb.gimp_image_set_active_layer(image,layer)
			
			wL = pdb.gimp_drawable_width(layer)
			hL = pdb.gimp_drawable_height(layer)

			new_w=(wL*percent)/100 
			new_h=(hL*percent)/100
			
			pdb.gimp_layer_scale(layer, new_w, new_h, local_origin) 			
 
		if scale==2:
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.gimp_layer_scale(layer, new_width, new_height, local_origin)
			
		if offset!=0:
			a,b = pdb.gimp_drawable_offsets(layer)
			pdb.gimp_image_set_active_layer(image,layer)
			if offset==1:
				a=0
				b=0
			pdb.gimp_layer_set_offsets(layer, a+offset_x, b+offset_y)
			
		
		if sel1:
			#pdb.gimp_context_set_background(background)	
			pdb.gimp_image_set_active_layer(image,layer)
			non_empty, x1, y1, x2, y2 = pdb.gimp_selection_bounds(image)
			pdb.gimp_drawable_edit_clear(layer)

		if sel3:
			pdb.gimp_image_set_active_layer(image,layer)
			#pdb.gimp_selection_bounds(image)		
			#pdb.gimp_image_select_item(image,0,layer)

			pdb.gimp_context_set_foreground(color_fg)
			pdb.gimp_drawable_edit_fill(layer,FILL_FOREGROUND)
			
		if sel4:
			pdb.gimp_image_set_active_layer(image,layer)
			#pdb.gimp_selection_bounds(image)		
			#pdb.gimp_image_select_item(image,0,layer)
			pdb.gimp_context_set_pattern(sel_pattern)
			pdb.gimp_drawable_edit_fill(layer,4)			

		if plugin==1:	# alienmap
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_alienmap2(image, layer, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, run_mode=vals)	
		if plugin==2:	# apply_canvas
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_apply_canvas(image, layer, 1, 1, run_mode=vals)			
		if plugin==3:	# Gaussian Blur
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_gauss(image, layer, 1, 1, 1, run_mode=vals)			
		if plugin==4:	# Selective Gaussian Blur
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_sel_gauss(image, layer, 1, 1, run_mode=vals)		
		if plugin==5:	# Motion Blur
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_mblur(image, layer, 1, 1, 1, 1, 1, run_mode=vals)		
		if plugin==6:	# bump_map
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_bump_map(image, layer, layer, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, run_mode=vals)
		if plugin==7:	# cartoon
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_cartoon(image, layer, 1, 1, run_mode=vals)
		if plugin==8:	# Channel Mixer...
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_colors_channel_mixer(image, layer, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, run_mode=vals)
		if plugin==9:	# Checkerboard...
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.plug_in_checkerboard(image, layer, 1, 1, run_mode=vals)
		if plugin==10:	# Rotate Colors
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_rotate_colormap(image, layer, run_mode=vals)

			
		if plugin==11:	# Convolution Matrix
			pdb.gimp_image_set_active_layer(image,layer)			
			pdb.plug_in_convmatrix(image, layer, 25, [0,0,0,0,0, 0,1,2,1,0, 0,2,4,2,0, 0,1,2,1,0, 0,0,0,0,0], False, 16, 0, 5, [True,True,True,True,True], 0, run_mode=vals)
		if plugin==12:	# Color to Alpha
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_colortoalpha(image, layer, (1,1,1), run_mode=vals)	
		if plugin==13:	# cubism
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_cubism(image, layer, 1, 1, 1, run_mode=vals)
		if plugin==14:	# Deinterlace
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_deinterlace(image, layer, 1, run_mode=vals)
		if plugin==15:	# displace
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_displace(image, layer, 1, 1, 1, 1, layer, layer, 1, run_mode=vals)
		if plugin==16:	# DoG
			pdb.gimp_image_set_active_layer(image,layer)			
			pdb.plug_in_dog(image, layer, 1, 1, 1, 1, run_mode=vals)		
		if plugin==17:	# edge
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_edge(image, layer, 1, 1, 1, run_mode=vals)
		if plugin==18:	# laplace
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_laplace(image, layer)
		if plugin==19:	# NEON (legacy)
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.plug_in_neon(image, layer, 1, 1, run_mode=vals)			
		if plugin==20:	# sobel
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_sobel(image, layer, 1, 1, 1, run_mode=vals)

			
		if plugin==21:	# emboss
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_emboss(image, layer, 1, 1, 1, 1, run_mode=vals)		
		if plugin==22:	# engrave
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_engrave(image, layer, 1, 1, run_mode=vals)
		if plugin==23:	# gimpressionist
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_gimpressionist(image, layer,"", run_mode=vals)
		if plugin==24:	# glasstile
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_glasstile(image, layer, 1, 1, run_mode=vals)
		if plugin==25:	# grid
			pdb.gimp_image_set_active_layer(image,layer)			
			pdb.plug_in_grid(image, layer, 1,1,1, (1,1,1), 1,1,1,1, (1,1,1), 1,1,1,1, (1,1,1), 1, run_mode=vals)			
		if plugin==26:	# illusion
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_illusion(image, layer, 1, 1, run_mode=vals)
		if plugin==27:	# Apply Lens
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_applylens(image, layer, 1, 1, 1, 1, run_mode=vals)
		if plugin==28:	# Lens Distortion
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_lens_distortion(image, layer, 1, 1, 1, 1, 1, 1, run_mode=vals)
		if plugin==29:	# maze
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_maze(image, layer, 1,  1, 1, 1, 1, 1, 1, run_mode=vals)	
		if plugin==30:	# mosaic
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_mosaic(image, layer, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, run_mode=vals)
			
		if plugin==31:	# newsprint
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_newsprint(image, layer, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, run_mode=vals)			
		if plugin==32:	# Noise hsv
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_hsv_noise(image, layer, 1, 1, 1, 1, run_mode=vals)			
		if plugin==33:	# Noise Hurl
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_randomize_hurl(image, layer, 1, 1, 1, 1, run_mode=vals)
		if plugin==34:	# Noise RGB
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_rgb_noise(image, layer, 1, 1, 1, 1, 1, 1, run_mode=vals)
		if plugin==35:	# noise	solid
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_solid_noise(image, layer, 1, 1, 1, 1, 1, 1, run_mode=vals)
		if plugin==36:	# noise-spread
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_spread(image, layer, 1, 1, run_mode=vals)
		if plugin==37: # oilify
			pdb.gimp_image_set_active_layer(image,layer)			
			pdb.plug_in_oilify(image, drawable, layer, 1, 1, run_mode=vals)	
		if plugin==38: # Photocopy
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.plug_in_photocopy(image, layer, 1, 1, 1, 1, run_mode=vals)			
		if plugin==39:	# pixelize
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_pixelize2(image, layer, 1, 1,  run_mode=vals)
		if plugin==40:	# plasma
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_plasma(image, layer, 1, 1,  run_mode=vals)
			
		if plugin==41:	# Polar Coordinates
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_polar_coords(image, layer, 1, 1, 1, 1, 1, run_mode=vals)
		if plugin==42:	# Red Eye Removal
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_red_eye_removal(image, layer, 1, run_mode=vals)	
		if plugin==43:	# ripple
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_ripple(image, layer, 1, 1, 1, 1, 1, 1, 1, run_mode=vals)
		if plugin==44:	# Shift
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_shift(image, layer, 1, 1, run_mode=vals)
		if plugin==45:	# sinus
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_sinus(image, layer, 1, 1, 1, 1, 1, 1, 1, (1,1,1), (1,1,1), 1, 1, 1, 1, run_mode=vals)
		if plugin==46:	# Softglow (legacy) Filters/Artistic
			pdb.gimp_image_set_active_layer(image,layer)			
			pdb.plug_in_softglow(image,layer, 1, 1, 1, run_mode=vals)
		if plugin==47:	# Threshold Alpha
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.plug_in_threshold_alpha(image, layer, 1, run_mode=vals)		
		if plugin==48:	# Unsharp Mask
			pdb.gimp_image_set_active_layer(image,layer)			
			pdb.plug_in_unsharp_mask(image, layer, 1, 1, 1, run_mode=vals)				
		if plugin==49:   # Value Propagate
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_vpropagate(image, layer, 1, 1, 1, 1, 1, 1, run_mode=vals)			
		if plugin==50:	# video
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_video(image, layer, 1, 1, 1, run_mode=vals)	
			
		if plugin==51:	# waves
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_waves(image, layer, 1, 1, 1, 1, 1, run_mode=vals)
		if plugin==52:	# wind
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_wind(image, layer, 1, 1, 1, 1, 1, run_mode=vals)
		if plugin==53:	# # WHIRL AND PINCH
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.plug_in_whirl_pinch(image, layer, 1, 1, 1, run_mode=vals)						
		if plugin==54:	# fimg_noise
			pdb.gimp_image_set_active_layer(image,layer)
			pdb.plug_in_fimg_noise(image, layer,"", run_mode=vals)
		if plugin==55:	# gmic_qt
			#pdb.plug_in_gmic_qt(image, drawable, input, output, command)
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_gmic_qt(image,layer, 1, 0, "", run_mode=vals)
		if plugin==56:	# mathmap
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.plug_in_mathmap(image, layer, 1, 1, 1, 1, run_mode=vals)	
		if plugin==57:
			pdb.gimp_image_set_active_layer(image,layer)		
			pdb.gimp_plugin_userfilter(image, layer, "1", "1", "1", "1", run_mode=vals)
			

		if Op==1:
			Operation = str('alien-map')
		if Op==2:
			Operation = str('apply-lens')
		if Op==3:		
			Operation = str('bilateral-filter')
		if Op==4:		
			Operation = str('bloom')		
		if Op==5:		
			Operation = str('box-blur')		
		if Op==6:		
			Operation = str('buffer-sink')		
		if Op==7:		
			Operation = str('c2g')		
		if Op==8:		
			Operation = str('cartoon')			
		if Op==9:		
			Operation = str('cubism')
		if Op==10:		
			Operation = str('edge')
			
		if Op==11:		
			Operation = str('edge-neon')		
		if Op==12:		
			Operation = str('edge-sobel')		
		if Op==13:		
			Operation = str('emboss')		
		if Op==14:		
			Operation = str('engrave rgb-clip')				
		if Op==15:		
			Operation = str('fattal02')
		if Op==16:		
			Operation = str('focus-blur')
		if Op==17:		
			Operation = str('fractal-trace')	
		if Op==18:		
			Operation = str('gamma')
		if Op==19:		
			Operation = str('gaussian-blur')
		if Op==20:		
			Operation = str('gaussian-blur-selective')
				
		if Op==21:		
			Operation = str('gblur-1d')
		if Op==22:		
			Operation = str('gray')
		if Op==23:		
			Operation = str('high-pass')
		if Op==24:		
			Operation = str('hue-chroma')
		if Op==25:		
			Operation = str('illusion')
		if Op==26:		
			Operation = str('invert-gamma')
		if Op==27:		
			Operation = str('invert-linear')			
		if Op==28:		
			Operation = str('lens-blur')
		if Op==29:		
			Operation = str('lens-flare')
		if Op==30:		
			Operation = str('long-shadow')
			
		if Op==31:		
			Operation = str('mantiuk06')
		if Op==32:		
			Operation = str('maze')
		if Op==33:		
			Operation = str('mblur')
		if Op==34:		
			Operation = str('mean-curvature-blur')
		if Op==35:		
			Operation = str('median-blur')
		if Op==36:		
			Operation = str('mirrors')
		if Op==37:		
			Operation = str('mosaic')			
		if Op==38:		
			Operation = str('motion-blur-circular')			
		if Op==39:		
			Operation = str('motion-blur-linear')
		if Op==40:		
			Operation = str('motion-blur-zoom')
			
		if Op==41:		
			Operation = str('newsprint')
		if Op==42:		
			Operation = str('noise-cie-lch')
		if Op==43:		
			Operation = str('noise-hurl')
		if Op==44:		
			Operation = str('noise-pick')
		if Op==45:		
			Operation = str('noise-reduction')
		if Op==46:		
			Operation = str('noise-rgb')
		if Op==47:		
			Operation = str('noise-slur')		
		if Op==48:		
			Operation = str('noise-spread')		
		if Op==49:		
			Operation = str('normal-map')	
		if Op==50:		
			Operation = str('oilify')
			
		if Op==51:		
			Operation = str('photocopy')	
		if Op==52:		
			Operation = str('pixelize')	
		if Op==53:		
			Operation = str('plasma')	
		if Op==54:		
			Operation = str('polar-coordinates')	
		if Op==55:		
			Operation = str('posterize')	
		if Op==56:		
			Operation = str('rectangle')	
		if Op==57:		
			Operation = str('scale-size')				
		if Op==58:		
			Operation = str('sepia')		
		if Op==59:		
			Operation = str('shift')
		if Op==60:		
			Operation = str('slic')
			
		if Op==61:		
			Operation = str('spherize')
		if Op==62:		
			Operation = str('stereographic-projection')
		if Op==63:		
			Operation = str('stress')
		if Op==64:		
			Operation = str('supernova')
		if Op==65:		
			Operation = str('texturize-canvas')
		if Op==66:		
			Operation = str('threshold')
		if Op==67:		
			Operation = str('tile-glass')	
		if Op==68:		
			Operation = str('tile-paper')	
		if Op==69:		
			Operation = str('tile-seamless')
		if Op==70:		
			Operation = str('unsharp-mask')
			
		if Op==71:		
			Operation = str('value-invert')
		if Op==72:		
			Operation = str('value-propagate')
		if Op==73:		
			Operation = str('variable-blur')
		if Op==74:		
			Operation = str('video-degradation')
		if Op==75:		
			Operation = str('vignette')
		if Op==76:		
			Operation = str('waterpixels')
		if Op==77:		
			Operation = str('watershed-transform')
		if Op==78:		
			Operation = str('waves')	
		if Op==79:		
			Operation = str('whirl-pinch')	
		if Op==80:		
			Operation = str('wind')
			
		if Op:
			pdb.gimp_image_set_active_layer(image,layer)			
			gegl_batch(image, layer, Operation)				
	

		if code_r==1:
			pdb.gimp_image_set_active_layer(image,layer)
			code = str(code)
			exec(code)
	
		if code_r==2:
			pdb.gimp_image_set_active_layer(image,layer)
			#dir = gimp.directory +  os.sep + 'plug-ins' + os.sep + 'Actions-Presets' +  os.sep
			dirpath = dir + str(actions_path()[load_name_actions])
			code = open(dirpath, "r")
			exec(code)
			code.close()

		if code_r==1 and code_save:	
			save_code(image, layer, code,name_preset)
	
		
		if mode==0: # NORMAL-LEGACY (0)
			pdb.gimp_layer_set_mode(layer, 0)
		if mode==1:  # DISSOLVE (1)
			pdb.gimp_layer_set_mode(layer, 1)
		if mode==2: # BEHIND-LEGACY (2)
			pdb.gimp_layer_set_mode(layer, 2)
		if mode==3:	# MULTIPLY-LEGACY (3)
			pdb.gimp_layer_set_mode(layer, 3)
		if mode==4:	# SCREEN-LEGACY (4)
			pdb.gimp_layer_set_mode(layer, 4)
		if mode==5:	# OVERLAY-LEGACY (5)
			pdb.gimp_layer_set_mode(layer, 5)
		if mode==6:	 # DIFFERENCE-LEGACY (6)
			pdb.gimp_layer_set_mode(layer, 6)
		if mode==7: # ADDITION-LEGACY (7)
			pdb.gimp_layer_set_mode(layer, 7)
		if mode==8:	# SUBTRACT-LEGACY (8)
			pdb.gimp_layer_set_mode(layer, 8)
		if mode==9:	# DARKEN-ONLY-LEGACY (9)
			pdb.gimp_layer_set_mode(layer, 9)
		if mode==10:	# LIGHTEN-ONLY-LEGACY (10)
			pdb.gimp_layer_set_mode(layer, 10)
		if mode==11:	 # HSV-HUE-LEGACY (11)
			pdb.gimp_layer_set_mode(layer, 11)
		if mode==12:	 # HSV-SATURATION-LEGACY (12)
			pdb.gimp_layer_set_mode(layer, 12)
		if mode==13:	# HSL-COLOR-LEGACY (13)
			pdb.gimp_layer_set_mode(layer, 13)
		if mode==14:	# HSV-VALUE-LEGACY (14)
			pdb.gimp_layer_set_mode(layer, 14)
		if mode==15:	# DIVIDE-LEGACY (15)
			pdb.gimp_layer_set_mode(layer, 15)
		if mode==16:	 # DODGE-LEGACY (16)
			pdb.gimp_layer_set_mode(layer, 16)
		if mode==17:	# BURN-LEGACY  (17)
			pdb.gimp_layer_set_mode(layer, 17)
		if mode==18:	# HARDLIGHT-LEGACY (18)
			pdb.gimp_layer_set_mode(layer, 18)
		if mode==19:	 # SOFTLIGHT-LEGACY (19)
			pdb.gimp_layer_set_mode(layer, 19)
		if mode==20:	 # GRAIN-EXTRACT-LEGACY (20)
			pdb.gimp_layer_set_mode(layer, 20)
		if mode==21:	 # GRAIN-MERGE-LEGACY (21)
			pdb.gimp_layer_set_mode(layer, 21)

	
		if mode==23:		# OVERLAY (23)
			pdb.gimp_layer_set_mode(layer, 23)			
		if mode==24:		# LCH-HUE (24)
			pdb.gimp_layer_set_mode(layer, 24)
		if mode==25:		# LCH-CHROMA (25)
			pdb.gimp_layer_set_mode(layer, 25)
		if mode==26:		# LCH-COLOR (26)
			pdb.gimp_layer_set_mode(layer, 26)
		if mode==27:		# LCH-LIGHTNESS (27)
			pdb.gimp_layer_set_mode(layer, 27)
		if mode==28:		# NORMAL (28),
			pdb.gimp_layer_set_mode(layer, 28)			
		if mode==29:		# BEHIND (29),
			pdb.gimp_layer_set_mode(layer, 29)			
		if mode==30:		# MULTIPLY (30),
			pdb.gimp_layer_set_mode(layer, 30)			
		if mode==31:		# SCREEN (31),
			pdb.gimp_layer_set_mode(layer, 31)			
		if mode==32:		# DIFFERENCE (32),
			pdb.gimp_layer_set_mode(layer, 32)			
		if mode==33:	# ADDITION (33),
			pdb.gimp_layer_set_mode(layer, 33)			
		if mode==34:	# SUBTRACT (34),
			pdb.gimp_layer_set_mode(layer, 34)			
		if mode==35:	# DARKEN-ONLY (35),
			pdb.gimp_layer_set_mode(layer, 35)			
		if mode==36:	# LIGHTEN-ONLY (36),
			pdb.gimp_layer_set_mode(layer, 36)			
		if mode==37:	# HSV-HUE (37),
			pdb.gimp_layer_set_mode(layer, 37)			
		if mode==38:	# HSV-SATURATION (38),
			pdb.gimp_layer_set_mode(layer, 38)			
		if mode==39:	# HSL-COLOR (39),
			pdb.gimp_layer_set_mode(layer, 39)			
		if mode==40:	# HSV-VALUE (40),
			pdb.gimp_layer_set_mode(layer, 40)			
		if mode==41:	# DIVIDE (41),
			pdb.gimp_layer_set_mode(layer, 41)			
		if mode==42:	# DODGE (42),
			pdb.gimp_layer_set_mode(layer, 42)			
		if mode==43:	# BURN (43),
			pdb.gimp_layer_set_mode(layer, 43)			
		if mode==44:	# HARDLIGHT (44),
			pdb.gimp_layer_set_mode(layer, 44)			
		if mode==45:	# SOFTLIGHT (45),
			pdb.gimp_layer_set_mode(layer, 45)			
		if mode==46:	# GRAIN-EXTRACT (46),
			pdb.gimp_layer_set_mode(layer, 46)			
		if mode==47:	# GRAIN-MERGE (47),	
			pdb.gimp_layer_set_mode(layer, 47)			
		if mode==48:	# VIVID-LIGHT (48)
			pdb.gimp_layer_set_mode(layer, 48)
		if mode==49:	# PIN-LIGHT (49)
			pdb.gimp_layer_set_mode(layer, 49)
		if mode==50:	# LINEAR-LIGHT (50)
			pdb.gimp_layer_set_mode(layer, 50)
		if mode==51:	# HARD-MIX (51)
			pdb.gimp_layer_set_mode(layer, 51)
		if mode==52:	# EXCLUSION (52)
			pdb.gimp_layer_set_mode(layer, 52)
		if mode==53:	# LINEAR-BURN (53)
			pdb.gimp_layer_set_mode(layer, 53)
		if mode==54:	# LUMA-DARKEN-ONLY (54)
			pdb.gimp_layer_set_mode(layer, 54)
		if mode==55:	# LUMA-LIGHTEN-ONLY (55)
			pdb.gimp_layer_set_mode(layer, 55)
		if mode==56:	# LUMINANCE (56)
			pdb.gimp_layer_set_mode(layer, 56)

			
		if xopacity==1:			
			pdb.gimp_layer_set_opacity(layer, opacity)
		
	count=0
	if rename==1: # " ADD THE NUMBER (Top->Down)\n TO THE NAME OF LAYER "
		for layer in apply_layers:
			oldname = layer.name
			rep_pos = oldname.find('- ')
			if rep_pos == -1:
				layer.name = oldname[rep_pos+1:]
			else: 
				layer.name = oldname[rep_pos+2:]
			layer.name = oldname[rep_pos+1:] 
			oldname = layer.name
			layer.name = str(count)+' - '+str(oldname)
			count+=1
	if rename==2: # " REPLACE THE NAME OF LAYERS\n THE NUMBER (Top->Down) "
		for layer in apply_layers:
			oldname = layer.name
			rep_pos = oldname.find('- ')		
			if rep_pos == -1:
				layer.name = oldname[rep_pos+1:]
			else: 
				layer.name = oldname[rep_pos+2:]
			layer.name = oldname[rep_pos+1:] 
			oldname = layer.name
			layer.name = str(count)
			count+=1
	if rename==3: # " ADD THE NUMBER (Bottom->Up)\n TO THE NAME OF LAYER "
		for layer in reversed(apply_layers):
			oldname = layer.name
			rep_pos = oldname.find('- ')		
			if rep_pos == -1:
				layer.name = oldname[rep_pos+1:]
			else: 
				layer.name = oldname[rep_pos+2:]
			layer.name = oldname[rep_pos+1:] 
			oldname = layer.name
			layer.name = str(count)+' - '+str(oldname)
			count+=1
	if rename==4: # " REPLACE THE NAME OF LAYERS\n THE NUMBER (Bottom->Up) "
		for layer in reversed(apply_layers):
			oldname = layer.name
			rep_pos = oldname.find('- ')		
			if rep_pos == -1:
				layer.name = oldname[rep_pos+1:]
			else: 
				layer.name = oldname[rep_pos+2:]
			layer.name = oldname[rep_pos+1:] 
			oldname = layer.name
			layer.name = str(count)
			count+=1
	if trans5 or trans6:	
		pdb.gimp_image_remove_layer(image, copy)

	if export==1:
		#for layer in apply_layers:
		pdb.gimp_image_set_active_layer(image,layer)
		exportLayers_arpl(image,layer,directory,namePattern)
			
	pdb.gimp_context_pop()
	pdb.gimp_image_undo_group_end(image)
	
	pdb.gimp_displays_flush()
	
hlp="	         *	For Layer Group	*	" \
+ "\n Working with layer groups can be confusing because 'Apply Actions' for 'Parent' affects 'Children' layers\n (error messages may appear: The element '...' cannot be modified because it is an element of a group item).\n Therefore, it is recommended to GROUP LAERS as the last operation (if You need to, of course...)."  \
+ "\n	         *	For Plugins & GEGL	*	" \
+ "\n This is the most important part of the plugin.\n Installing additional plugins is not necessary - but it does give You thousands of new filters to use \n (and the number of combinations is practically unlimited)." \
+ "\n	         *	For the UserFilter and MathMap	*	" \
+ "\n In RUN_INTERACTIVE mode, the filter must be selected in the main window \n (selection from a separate submenu will not work). " \
+ "\n	         *	For CUSTOM CODE	*	" \
+ " \n Depending on the procedure used (plugins.py or script.scm), the multi-layer operation will fail when \n the plug-in creates: multiple layers (without the 'Merge' option) or a new image or resizes the image. "  \
+ "\n	         *	For COLOR	*	" \
+ "\n The format of saving curves has already changed several times.\n For this plugin, 'Curves' saved in Gimp-2.8 or later should work fine."

dir = gimp.directory +  os.sep + 'plug-ins' +  os.sep + 'Actions-Presets' +  os.sep

dirf = gimp.directory +  os.sep + 'curves' +  os.sep + "*"
register(
	"python_fu_MrQ_Actions_Run_Plugin_on_Layers",
	"You can combine multiple actions (in sequence) in one run, but not all. If the actions cannot be merged or the workflow requires a different order, do what you can and show the plug-in again (Shift + Ctrl + F) and apply another (but don't forget to check which ones are active). ",
	"Actions Run Plugin (or code) on Layers Ver.1.2",
	"MareroQ",
	"GPL",
	"2021",
	"Actions RPL...",
	"*",
	[
	(PF_IMAGE, "image",       "Input image", None),
	(PF_DRAWABLE, "layer", "Input drawable", None),	
	(PF_OPTION, "layergroup"," Apply Actions to:\n (select layers) ", 5, [" ALL LAYERS "," VISIBLE LAYERS "," NOT-VISIBLE LAYERS "," LINKED LAYERS "," NOT-LINKED LAYERS "," ACTIVE ONLY "," ACTIVE AND ABOVE ", " ACTIVE AND BELOW "]),
	(PF_OPTION, "rev", " Target Result Layers: ", 0, [" DON'T USE ", " VISIBLE LAYER ", " NOT-VISIBLE LAYER ", " LINKED LAYER ", " NOT-LINKED LAYER "," GROUP IN ONE "," GROUP INDIVIDUALLY "," UNGROUP GROUP LAYER "," RAISES LAYER TO TOP\n ('Reverse') "," RAISES LAYER TO BOTTOM\n ('Reverse')"," CLEAR LAYERS "," REMOVE LAYERS ", " REVERSE VISIBILITY \n (visible → invisible | invisible → visible)"," SHUFFLE LAYERS "]),	
	(PF_OPTION, "mode", " Layer-Mode: ", 22, 
	[" NORMAL-LEGACY (0) ",
	" DISSOLVE (1) ",
	" BEHIND-LEGACY (2) ",
	" MULTIPLY-LEGACY (3) ",
	" SCREEN-LEGACY (4) ",
	" OVERLAY-LEGACY (5) ",
	" DIFFERENCE-LEGACY (6) ",
	" ADDITION-LEGACY (7) ",
	" SUBTRACT-LEGACY (8) ",
	" DARKEN-ONLY-LEGACY (9) ",
	" LIGHTEN-ONLY-LEGACY (10) ",
	" HSV-HUE-LEGACY (11) ",
	" HSV-SATURATION-LEGACY (12) ",
	" HSL-COLOR-LEGACY (13) ",
	" HSV-VALUE-LEGACY (14) ",
	" DIVIDE-LEGACY (15) ",
	" DODGE-LEGACY (16) ",
	" BURN-LEGACY  (17) ",
	" HARDLIGHT-LEGACY (18) ",
	" SOFTLIGHT-LEGACY (19) ",
	" GRAIN-EXTRACT-LEGACY (20) ",
	" GRAIN-MERGE-LEGACY (21) ",
	" DON'T USE\n"u'\u2191'" Legacy  -  New (default) "u'\u2193'" ", 
	" OVERLAY (23) ",
	" LCH-HUE (24) ",
	" LCH-CHROMA (25) ",
	" LCH-COLOR (26) ",
	" LCH-LIGHTNESS (27) ", 
	" NORMAL (28) ",
	" BEHIND (29) ",
	" MULTIPLY (30) ",
	" SCREEN (31) ",
	" DIFFERENCE (32) ",
	" ADDITION (33) ",
	" SUBTRACT (34) ",
	" DARKEN-ONLY (35) ",
	" LIGHTEN-ONLY (36) ",
	" HSV-HUE (37) ",
	" HSV-SATURATION (38) ",
	" HSL-COLOR (39) ",
	" HSV-VALUE (40) ",
	" DIVIDE (41) ",
	" DODGE (42) ",
	" BURN (43) ",
	" HARDLIGHT (44) ",
	" SOFTLIGHT (45) ",
	" GRAIN-EXTRACT (46) ",
	" GRAIN-MERGE (47) ",
	" VIVID-LIGHT (48) ",
	" PIN-LIGHT (49) ",
	" LINEAR-LIGHT (50) ",
	" HARD-MIX (51) ",
	" EXCLUSION (52) ",
	" LINEAR-BURN (53) ",
	" LUMA-DARKEN-ONLY (54) ",
	" LUMA-LIGHTEN-ONLY (55) ",
	" LUMINANCE (56) ",
	]),	

	(PF_TOGGLE,  "xopacity", " Change the opacity: ", 0,[("Don't Use "),("Action Active")]),
	(PF_SPINNER, "opacity",  "   - opacity (%): ", 100, (0, 100, 1)),
	
	(PF_TOGGLE,  "cocu1", " COPY & PASTE TOP to DOWN: ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':1}),
	(PF_TOGGLE,  "cocu2", " COPY & PASTE REVERSED: ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':1}),
	(PF_TOGGLE,  "cocu3", " COPY & PASTE ABOVE CHOOSEN LAYERS: ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':1}),
	(PF_TOGGLE,  "cocu4", " COPY & PASTE BELOW CHOOSEN LAYERS: ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':1}),
	(PF_TOGGLE,  "cocu5", " COPY MULTIPLE TIMES: ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':1}),
	(PF_SPINNER, "iter",  "  - number of copies: ", 0, (0, 5000, 1),{'newLine':'1','tab':1}),
	(PF_TOGGLE,  "cocu6", " OPEN AS LAYERS IN NEW IMAGE: ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':1}),
	(PF_TOGGLE,  "cocu7", " OPEN LAYERS AS NEW IMAGES: ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':1}),

	(PF_TOGGLE,  "mir1", " CENTERED : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':2}),
	(PF_TOGGLE,  "mir2", " ALIGN TOP : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':2}),
	(PF_TOGGLE,  "mir3", " ALIGN BOTTOM : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':2}),
	(PF_TOGGLE,  "mir4", " ALIGN RIGHT : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':2}),
	(PF_TOGGLE,  "mir5", " ALIGN LEFT : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':2}),
	
	(PF_TOGGLE,  "mir6", " FLIP HORIZONTAL "u'\u21e8'"", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':2}),
	(PF_TOGGLE,  "mir7", " FLIP VERTICAL "u'\u21e9'"", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':2}),
	(PF_TOGGLE,  "mir8", " ROTATE-90° : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':2}),
	(PF_TOGGLE,  "mir9", " ROTATE-180° : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':2}),
	(PF_TOGGLE,  "mir10", " ROTATE-270° : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':2}),	

	(PF_TOGGLE,  "trans1", " ADD ALPHA CHANNEL : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':3}),
	(PF_TOGGLE,  "trans2", " REMOVE ALPHA CHANNEL :\n (blend transparent pixels\n against the background color)", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':3}),
	(PF_TOGGLE,  "trans3", " LOCK THE ALPHA : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':3}),
	(PF_TOGGLE,  "trans4", " UNLOCK THE ALPHA : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':3}),
	(PF_TOGGLE,  "trans5", " OVERLAY\n (active layer on chosen) : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':3}),
	(PF_TOGGLE,  "trans6", " COMBINE\n (active layer under the chosen) : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':3}),
	(PF_TOGGLE,  "trans7", " ALPHA TO PATH : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':3}),
	(PF_TOGGLE,  "trans8", " ALPHA TO SELECTION : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':3}),
	(PF_TOGGLE,  "trans9", " SEMIFLATTEN\n (Replace partial transparency\n  with the current BG color) : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':3}),
	
	(PF_TITLE,  "color_info"  , "<b>				This only valid on RGB color images</b>",(0.0,0.0), None  ,{'iniCol':1,'cols':2,'tab':4}),	
	(PF_TOGGLE,  "color_p", " APPLY CURVES BY SAVED PRESETS : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':4}),
	(PF_OPTION,  "name_p",  " - Color Curves Presets:", 0, local_get_color_curves_presets_names(),{'newLine':'1','tab':4}),
	(PF_TOGGLE,  "color_f", " APPLY CURVES BY FILE : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':4}),
	(PF_FILENAME, "name_f", " - Curve File to Load :\n   (default from dir 'curves') ", dirf,0,{'newLine':'1','tab':4}),
	(PF_TOGGLE,  "colorize", " COLORIZE \n (as grayscale image seen\n through a colored glass) : ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':4}),
	(PF_COLOR,  "color_hsl"    , "Color [hsl]:", (130, 148, 213),0,{'newLine':'1','tab':4}),
	(PF_TOGGLE,  "color_rem", " Remove Color :\n (for use transparent\n 'add alpha chanel') ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':4}),
	(PF_COLOR,  "color_rgb"    , " - Color to Remove:", (130, 148, 213),0,{'newLine':'1','tab':4}),

	

	(PF_OPTION,  "sel", " Scale: ", 0, ["Don't Use ","Scale by Percent (%)","Scale by new width/height"],{'newLine':'1','tab':5}),	
	(PF_OPTION,  "interpolation", " Interpolation: ", 3, [" NONE ", " LINEAR ", " CUBIC ", " NOHALO ", " LOHALO "],{'newLine':'1','tab':5}),
	(PF_TOGGLE,  "local_origin", " Scale from: ", 1,[("Image origin"),("Center of the layer ")],{'newLine':'1','tab':5}),	
	(PF_SPINNER, "percent", "Percent (%):", 50, (1, 1000, 1),{'newLine':'1','tab':5}),
	(PF_SPINNER, "new_width", "New Width (px):", 1, (1, 5000, 1),{'newLine':'1','tab':5}),
	(PF_SPINNER, "new_height", "New Height (px):", 1, (1, 5000, 1),{'newLine':'1','tab':5}),
	(PF_OPTION,  "offset", " Offset: ", 0, ["Don't Use ","Relative to the Image Origin\n(from coordinates: x=0,y=0)","Absolute Layer Shift"],{'newLine':'1','tab':5}),
		
	(PF_SPINNER, "offset_x", "Offset in x direction:", 0, (-5000, 5000, 1),{'newLine':'1','tab':5}),	
	(PF_SPINNER, "offset_y", "Offset in y direction:", 0, (-5000, 5000, 1),{'newLine':'1','tab':5}),	

	(PF_TITLE  ,  "sel_info"  , "<b>\n   ( if layer does not have an alpha channel, cleared pixels will be set to BG color )</b>",(0.0,0.0), None  ,{'iniCol':1,'cols':2,'tab':6}),	
	(PF_TOGGLE,  "sel1", " Clear selected area? ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':6}),
	(PF_TOGGLE,  "sel2", " Save to Chanel? ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':6}),
	(PF_TOGGLE,  "sel3", " Fill in with color? ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':6}),
	(PF_COLOR,  "color_fg"    , "  - Color ", (130, 148, 213),0,{'newLine':'1','tab':6}),		
	(PF_TOGGLE,  "sel4", " Fill in a pattern  ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':6}),
	(PF_PATTERN,  "sel_pattern", "  - Pattern :", "Crack",0,{'newLine':'1','tab':6}),
	
	(PF_TITLE,  "plugin_info"  , "<b>							PLUGINS\n\n     For GIMP 2.10 it is necessary to install 45 old plugins with GIMP 2.8 (are marked #)</b>\n    MultiFiler: 1.FelimageNoise, 2.G'MIC-QT, 3.MATHMAP , 4.USER FILTER - You should install additionally.\n		Hint: first use the effect INTERACTIVE on one layer (adjust the required parameters),\n	undo action (Ctr+Z) and then (for selected: Apply Layers) use the mode RUN_WITH_LAST_VALS.",(0.0,0.0), None  ,{'iniCol':1,'cols':2,'tab':7}),	
	(PF_OPTION, "vals", " Run-mode for PLUGIN: ", 1, [" INTERACTIVE "," RUN_WITH_LAST_VALS "],{'newLine':'1','tab':7}),		
	(PF_OPTION, "plugin", " Choose a PLUGIN: ", 0, [" DON'T USE ",
	" # ALIEN MAP # \n → Colors/Map/Alien Map ",									#  1 Colors/Map/Alien Map...
	" # APPLY CANVAS # \n → Filters/Artistic/Apply Canvas ",						#  2 Filters/Artistic/Apply Canvas...
	" # Blur GAUSSIAN BLUR # \n → Filters/Blur/Gaussian Blur ", 					#  3 Filters/Blur/Gaussian Blur...
	" # Blur SELECTIVE GAUSSIAN BLUR # \n → Filters/Blur/Selective Gaussian Blur ",	#  4 Filters/Blur/Selective Gaussian Blur...
	" # Blur MOTION BLUR # \n → Filters/Blur/Motion Blur ",							#  5 Filters/Blur/Motion Blur...
	" # BUMP MAP # \n → Filters/Map/Bump Map ",										#  6 Filters/Map/Bump Map...
	" CARTOON\n → Filters/Artistic/Cartoon (legacy) ",							    #  7 Filters/Artistic/Cartoonizer...	
	" # CHANNEL MIXER # \n → Colors/Components/Channel Mixer ",						#  8 Colors/Components/Channel Mixer...
	" CHECKERBOARD\n → Filters/Render/Pattern/Checkerboard (legacy) ",              #  9 Checkerboard Filters/Render/Pattern
	" # COLORS ROTATE # \n → Colors/Map/Rotate Colors",								# 10 Colors/Map/Rotate Colors...
	
	" # CONVOLUTION MATRIX # \n → Filters/Generic",									# 11 Filters/Generic	
	" # COLOR TO ALPHA # \n → Layer/Transparency/Color to Alpha ", 				    # 12 Layer/Transparency/Color to Alpha...
	" # CUBISM # \n → Filters/Artistic/Cubism ",									# 13 Filters/Artistic/Cubism...
	" # DEINTERLACE #\n → Filters/Enhance/Deinterlace ",                    		# 14 Filters/Enhance/Deinterlace...
	" # DISPLACE #\n → Filters/Map/Displace ", 										# 15 Filters/Map/Displace...
	" Edge-Detect DIFFERENCE OF GAUSSIANS (legacy) \n → (DoG) Filters/Edge-Detect", # 16 Filters/Edge-Detect/Difference of Gaussians (legacy)	
	" # Edge-Detect EDGE # \n → Filters/Edge-Detect/Edge ",               		    # 17 Filters/Edge-Detect/Edge...
	" # Edge-Detect LAPLACE #\n → Filters/Edge-Detect/Laplace", 					# 18 Filters/Edge-Detect/Laplace...
	" # NEON (legacy) #\n → Filters/Edge-Detect/Neon (legacy)",                     # 19 Filters/Edge-Detect/Neon (legacy)
	" # Edge-Detect SOBEL #\n → Filters/Edge-Detect/Sobel ", 	            		# 20 Filters/Edge-Detect/Sobel...
	
	" EMBOSS\n → Filters/Distorts/Emboss (legacy) ",			                	# 21 Filters/Distorts/Emboss (legacy)...
	" # ENGRAVE # \n → Filters/Distorts/Engrave ",									# 22 Filters/Distorts/Engrave...
	" GIMPRESSIONIST\n → Filters/Artistic/GIMPressionist ",							# 23 Filters/Artistic/GIMPressionist...
	" # GLASS TILE #\n → Filters/Artistic/Glass Tile ",								# 24 Filters/Artistic/Glass Tile...
	" GRID (legacy)\n → Filters/Render/Pattern",									# 25 Filters/Render/Pattern
	" # ILLUSION #\n → Filters/Map/Illusion ",                       				# 26 Filters/Map/Illusion...
	" # LENS EFFECT #\n → Filters/Distorts/Apply Lens ",                    		# 27 Filters/Distorts/Apply Lens...
	" # LENS DISTORTION #\n → Filters/Distorts/Lens Distortion ",               	# 28 Filters/Distorts/Lens Distortion...
	" # MAZE #\n → Filters/Render/Pattern/Maze ",	                        		# 29 Filters/Render/Pattern/Maze...
	" # MOSAIC #\n → Filters/Distorts/Mosaic ",										# 30 Filters/Distorts/Mosaic...
	
	" # NEWSPRINT #\n → Filters/Distorts/Newsprint ",								# 31 Filters/Distorts/Newsprint...
	" # Noise HSV #\n → Filters/Noise/HSV Noise ",                      			# 32 Filters/Noise/HSV Noise...
	" # Noise HURL #\n → Filters/Noise/Hurl ",                     					# 33 Filters/Noise/Hurl...
	" # Noise RGB #\n → Filters/Noise/RGB Noise ",                      			# 34 Filters/Noise/RGB Noise...
	" # Noise SOLID #\n → Filters/Render/Clouds/Solid Noise ",                    	# 35 Filters/Render/Clouds/Solid Noise...
	" # Noise SPREAD #\n → Filters/Noise/Spread ",									# 36 Filters/Noise/Spread...
	" # OILIFY # #\n → Filters/Artistic/Oilify",									# 37 Filters/Artistic/Oilify...
	" PHOTOCOPY (legacy)\n → Filters/Artistic",										# 38 Filters/Artistic
	" # PIXELIZE #\n → Filters/Blur/Pixelize ",										# 39 Filters/Blur/Pixelize...
	" # PLASMA #\n → Filters/Render/Clouds/Plasma ",                         		# 40 Filters/Render/Clouds/Plasma...
	
	" # POLAR COORDINATES #\n → Filters/Distorts/Polar Coordinates ",             	# 41 Filters/Distorts/Polar Coordinates...
	" # RED EYE REMOVAL #\n → Filters/Enhance/Red Eye Removal ",                	# 42 Filters/Enhance/Red Eye Removal...
	" # RIPPLE #\n → Filters/Distorts/Ripple ",                         			# 43 Filters/Distorts/Ripple...
	" # SHIFT #\n → Filters/Distorts/Shift ",                          				# 44 Filters/Distorts/Shift...
	" # SINUS #\n → Filters/Render/Pattern/Sinus ",                          		# 45 Filters/Render/Pattern/Sinus...	
	" SOFTGLOW (legacy)\n → Filters/Artistic", 									    # 46 Filters/Artistic
	" # Threshold Alpha #\n → Layer/Transparency/Threshold Alpha",					# 47 Threshold Alpha
	" # UNSHARP MASK #\n → Filters/Enhance/Unsharp Mask", 						    # 48 Filters/Enhance/Unsharp Mask...
	" # VALUE PROPAGATE #\n → Filters/Distorts/Value Propagate ",                	# 49 Filters/Distorts/Value Propagate...
	" # VIDEO #\n → Filters/Distorts/Video ",                          				# 50 Filters/Distorts/Video...
	
	" # WAVES #\n → Filters/Distorts/Waves ",                          				# 51 Filters/Distorts/Waves...
	" # WIND #\n → Filters/Distorts/Wind ",     									# 52 Filters/Distorts/Wind...
	" # WHIRL AND PINCH #\n → Filters/Distorts/Whirl and Pinch",					# 53 Filters/Distorts/Whirl and Pinch...
	" 1. FelimageNoise\n → Filters/Render/Pattern ",								# 54
	" 2. G'MIC\n Filters/→  G'MIC-Qt ",												# 55
	" 3. MATHMAP\n → Filters/Generic/MathMap",										# 56
	" 4. USER FILTER (Filter Factory Converter)\n → Filters/Generic/User Filter", 	# 57
	],{'newLine':'1','tab':7}),
	(PF_TITLE,  "gegl_info"  , "<b>	------------------------------------------------------------------------------------------------------\n							 GEGL\n						(only default values)</b>",(0.0,0.0), None  ,{'iniCol':1,'cols':2,'tab':7}),	
	
	(PF_OPTION, "Op", "Operation", 0, [" DON'T USE ",
	"Alien-map", "Apply-lens", "Bilateral-filter" ,"Bloom", "Box-blur", "Buffer-sink", "c2g","Cartoon","Cubism","Edge",
	"Eedge-neon","Edge-sobel","Emboss","Engrave","Fattal02", "Focus-blur", "Fractal-trace","Gamma", "Gaussian-blur", "Gaussian-blur-selective",
	"Gblur-1d", "Gray", "High-pass","Hue-chroma","Illusion","Invert-gamma","Invert-linear",
	"Lens-blur","Lens-flare","Long-shadow","Mantiuk06","Maze","mblur","Mean-curvature-blur","Median-blur","Mirrors","Mosaic",
	"Motion-blur-circular","Motion-blur-linear","Motion-blur-zoom","Newsprin","Noise-cie-lch","Noise-hurl","Noise-pick","Noise-reduction","Noise-rgb","Noise-slur",
	"Noise-spread","Normal-map","Oilify","Photocopy","Pixelize","Plasma","Polar-coordinates","Posterize","Rectangle","Scale-size",
	"Sepia","Shift","Slic","Spherize","Stereographic-projection","Stress","Supernova","Texturize-canvas","Threshold","Tile-glass",
	"Tile-paper","Tile-seamless","Unsharp-mask","Value-invert","Value-propagate","Variable-blur","Video-degradation","Vignette","Waterpixels","Watershed-transform",
	"Waves","Whirl-pinch","Wind"		
	],{'newLine':'1','tab':7}),
	
	(PF_OPTION,   "code_r", " Run Custom Code?: ", 0, [" Don't Use "," 1. Code to Run "," 2. Load Actions from Presets "],{'newLine':'1','tab':8}),	
	(PF_TITLE  ,  "code_info"  , "<b>		----------------------	1. Code to Run	----------------------\n				Separate code [pdb or gegl]\n				 - semicolon ';' or \n				 - use 'Enter' to new line.</b>",(0.0,0.0), None  ,{'iniCol':1,'cols':2,'tab':8}),		
	(PF_TEXT,     "code", " 1. Code to Run:\n (write or paste) ","gegl_batch(image, layer,'tile-seamless')",0,{'newLine':1,'tab':8}),
	(PF_TOGGLE,   "code_save", " - Save 'Code to Run'? ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':8}),
	(PF_TITLE  ,  "code_info_name"  , "<b> 		Code is saved/load in the home directory: .../plug-ins/Actions-Presets/. \n Don't forget to enter a unique preset name (to avoid being overwritten - without warning!).</b>",(0.0,0.0), None  ,{'iniCol':1,'cols':2,'tab':8}),	
	(PF_STRING,   "name_preset",  ' - Name Presets:', "Your name ?",0,{'newLine':1,'tab':8}),
	(PF_TITLE  ,  "load_info"  , "<b>		----------------	2. Load Actions from File	----------------</b>",(0.0,0.0), None  ,{'iniCol':1,'cols':2,'tab':8}),	
	#(PF_FILENAME, "file", " 2. File to Load : ", dir,0,{'newLine':'1','tab':8}),
	(PF_OPTION,  "load_name_actions",  "Actions Presets:",0, actions_path(),{'newLine':'1','tab':8}),
	
	(PF_TOGGLE,  "canva1", " AUTOCROP: ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':9}),	
	(PF_TOGGLE,  "canva2", " LAYERS TO IMAGE SIZE: ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':9}),
	(PF_TOGGLE,  "canva3", " IMAGE RESIZE TO LAYERS: ", 0,[("Don't Use "),("Action Active")],{'newLine':'1','tab':9}),
	
	(PF_OPTION, "rename", " Change layers-name: ", 0, [" DON'T USE ", " ADD THE NUMBER (Top->Down)\n TO THE NAME OF LAYER ", " REPLACE THE NAME OF LAYERS\n THE NUMBER (Top->Down) "," ADD THE NUMBER (Bottom->Up)\n TO THE NAME OF LAYER "," REPLACE THE NAME OF LAYERS\n THE NUMBER (Bottom->Up) "],{'newLine':'1','tab':10}),

	(PF_TOGGLE,  "export", "  Export layers: : ", 0,[("Don't Use "),("Export")],{'newLine':1,'tab':11}),	
	(PF_DIRNAME,    "directory",    'Directory', 0,  '.',{'newLine':1,'tab':11}),
	(PF_STRING,     'namePattern',  'Layer name', '{imageName}-{numDown1:03d}.png',0,{'newLine':1,'tab':11}),
	
	(PF_TAB    , "ptab"    , "tabs" , 0, (_('Source\n    &  \nTarget'),_('Copy\n   &\nOpen'),_('Align\n Flip\nRotate'),_('Alpha'),_('Color'),_('Scale\n   &\nOffset'),_('Selections'),_('Plugins\n    &    \n  GEGL  '),_("CUSTOM\n  CODE "),_('Autocrop\n RE-SIZE'),_('Change\n Layers\n  name '),_('Export\nLayers'),_('Help')) ),	
	(PF_TITLE  , "help"  , hlp ,(10,10), None           ,{'tab':12}),	
	],
	[],
	python_fu_MrQ_Actions_Run_Plugin_on_Layers, menu="<Image>/Layer/")

main()
